#! /usr/bin/env ruby

require 'optparse'
require 'pp'
require 'set'
require 'rest_client'
require 'base64'
require 'json'

module WTBuildHelpers 
end

module WTBuildHelpers::JIRA
    Options = Struct.new(:commit_range, :site, :username, :password, :build)

    class JiraInfo
        attr_reader :fixed_in_build_field_id
        attr_reader :build_complete_transition_id

        attr_reader :client
        
        def initialize(site, username, password)
            @site = site
            @username = username
            @password = password
            
            auth_token = Base64.strict_encode64("#{username}:#{password}")
            base_url = File.join(site, '/rest/api/2')
            headers = { "Authorization" => "Basic #{auth_token}",
                        "Content-Type" => "application/json", }
            @client = RestClient::Resource.new(base_url, :headers => headers)
        end
        
        def update_fixed_in_build()
            # Find the Fixed In Build field
            begin
                response = @client['field'].get()

                fields = JSON.parse(response.body);
                fixed_field = fields.find do |field|
                    field["name"] == "Fixed In Build"
                end
            
                if fixed_field
                    @fixed_in_build_field_id = fixed_field["id"]
                else
                    puts "Could not find 'Fixed In Build' field on supplied JIRA instance. Are you using WillowTree's JIRA? Does your user have access to the project?"
                end
            rescue RestClient::Exception => e
                puts "Error code #{e.response.code} attempting to find Fixed In Build field"
            end 
        end
        
        def update_build_complete_transition(issue_key)
            begin
                fields_resource = "issue/#{issue_key}/transitions"
            
                response = @client[fields_resource].get
            
                response_json = JSON.parse(response.body)
                transitions = response_json["transitions"]
                build_complete_transition = transitions.find do |transition|
                    transition["name"] == "Build Complete"
                end
            
                if build_complete_transition
                    @build_complete_transition_id = build_complete_transition["id"]
                end
            rescue RestClient::Exception => e
                puts "Error code #{e.response.code} attempting to find Build Complete transition"
            end
        end
    end

    class Parser
        def self.parse(options)
            args = Options.new()
       
            opt_parser = OptionParser.new do |opts|
                opts.banner = "Usage: jira_update_from_build [options]"
            
                opts.on("-c", "--commits COMMIT_RANGE", "Specify a commit range") do |commit_range|
                    args.commit_range = commit_range
                end
                
                opts.on("-u", "--username USERNAME", "Specify the user to connecto to JIRA") do |username|
                    args.username = username
                end
                
                opts.on("-p", "--password PASSWORD", "Specify the password to connect to JIRA") do |password|
                    args.password = password
                end
                
                opts.on("-s", "--site SITE", "Specify the base location of the JIRA instance") do |site|
                    site.slice!("http://")
                    site.insert(0, "https://") if not site.start_with?("https://")
                    args.site = site
                end
                
                opts.on("-b", "--build BUILD", "The build number for which this script is executing") do |build|
                    args.build = build
                end
                
                opts.on_tail("-h", "--help", "Show this message") do
                    puts opts
                    exit
                end
            end
        
            opt_parser.parse!(options)
            return args   
        end
    end

    def self.find_jira_issues(logs)
        issue_set = Set.new
        logs.each do |log|
            issues = log.scan(/[A-Za-z][A-Za-z]+-[1-9][0-9]*/)
            issue_set.merge(issues)
        end
        
        return issue_set
    end

    def self.get_git_history(commit_range)
        puts "Getting git history for #{commit_range}"
        happyface = "\u263A".encode('utf-8')
        extra_params = commit_range.index('..') == nil ? "-n 1" : ""
        git_log = `git --no-pager log --format="%B#{happyface}" #{commit_range} #{extra_params}`
        puts(git_log)

        #parse out the log
        logs = git_log.split(happyface)
        
        return logs
    end

    def self.jira_update(jira_info, issues, build)
        if !jira_info.fixed_in_build_field_id
            jira_info.update_fixed_in_build
        end
        
        # Something's wrong. We're not on our JIRA
        return unless jira_info.fixed_in_build_field_id
        
        issues.each do |issue_key|
            #try to perform the transition
            if !jira_info.build_complete_transition_id
                jira_info.update_build_complete_transition(issue_key)
            end
            
            #still nil? Can't perform this transition
            if !jira_info.build_complete_transition_id
                puts "Could not update issue #{issue_key}: could not find transition"
                next
            end
            
            params = {
                transition: {
                    id: jira_info.build_complete_transition_id
                },
                fields: {
                    jira_info.fixed_in_build_field_id => build
                }
            }

            puts "Updating issue #{issue_key}"
            begin
                issue_response = jira_info.client["issue/#{issue_key}/transitions"].post(JSON.generate(params))
            rescue RestClient::Exception => e
                puts "Error updating issue #{issue_key}: Error code #{e.response.code}"
            end
        end
    end

    def self.pull_travis_options(options)
        puts "Pulling parameters from Travis-CI environment variables where needed...."
        options.commit_range = ENV["TRAVIS_COMMIT_RANGE"] if options.commit_range == nil
        options.build = ENV["TRAVIS_BUILD_NUMBER"] if options.build == nil
    end

    def self.pull_team_city_options(options)
        puts "Pulling parameters from TeamCity environment variables where needed...."
        options.commit_range = ENV["TEAMCITY_COMMIT_RANGE"] if options.commit_range == nil
        options.build = ENV["BUILD_NUMBER"] if options.build == nil
    end

    def self.pull_common_options(options)
        options.site = ENV["JIRA_URL"] if options.site == nil
        options.username = ENV["JIRA_USERNAME"] if options.username == nil
        options.password = ENV["JIRA_PASSWORD"] if options.password == nil
    end

    def self.validate_options(options)
        valid_options = true
        if options.commit_range == nil 
            puts "Commit range is required. None specified."
            puts "You can specify this in TeamCity by running 'teamcity_fetch_git_range' before this step." if ENV["TEAMCITY_VERSION"]
            valid_options = false
        end
        if options.site == nil
            puts "Site parameter is required, or use JIRA_URL environment variable on build servers"
            valid_options = false
        end
        if options.username == nil
            puts "No username for JIRA provided. Use JIRA_USERNAME environment variable on build servers"
            valid_options = false
        end
        if options.password == nil
            puts "No password for JIRA provided. Use JIRA_PASSWORD environment variable on build servers"
            valid_options = false
        end

        valid_options
    end

    def self.print_help()
        Parser.parse(["--help"])
    end

    def self.commandline_jira_update(args)
        options = args ? Parser.parse(ARGV) : Options.new

        if ENV["TRAVIS"]
            pull_travis_options(options)
        # Detect running on TeamCity for common prarams 
        elsif ENV["TEAMCITY_VERSION"]
            pull_team_city_options(options)
        end
        pull_common_options(options)

        if validate_options(options)    
            logs = get_git_history(options.commit_range)
            issue_set = find_jira_issues(logs)
            
            if issue_set.size > 0
                jira_info = JiraInfo.new options.site, options.username, options.password 
                jira_update(jira_info, issue_set, options.build)
            end
        end
    end
end

if __FILE__ == $0
    WTBuildHelpers::JIRA::commandline_jira_update(ARGV)
end