#!/usr/bin/ruby

require 'optparse'
require 'base64'
require 'json'
require 'rest_client'
require 'pp'
require 'uri'

module WTBuildHelpers
end

module WTBuildHelpers::TeamCity
    Options = Struct.new(:url, :build_type_id, :username, :password, :export)

    class OptionsParser
        def self.parse(options)
            args = Options.new()
            args.export = false
       
            opt_parser = OptionParser.new do |opts|
                opts.banner = "Usage: #{__FILE__} [options]"
            
                opts.on("-s", "--url URL", "Specify the URL for TeamCity") do |url|
                    url.slice!("http://")
                    url.slice!("https://")
                    url.insert(0, "https://")
                    args.url = url
                end
                
                opts.on("-u", "--username USERNAME", "Specify the user to connecto to TeamCity") do |username|
                    args.username = username
                end
                
                opts.on("-p", "--password PASSWORD", "Specify the password to connect to TeamCity") do |password|
                    args.password = password
                end
                
                opts.on("-t", "--build_type BUILD_TYPE", "The build_type_id for the running build from TeamCity") do |build_type_id|
                    args.build_type_id = build_type_id
                end

                opts.on("-e", "--export", "Output range for use as a parameter to the 'export' bash command") do 
                    args.export = true
                end

                opts.on_tail("-h", "--help", "Show this message") do
                    puts opts
                    exit
                end
            end
        
            opt_parser.parse!(options)
            return args   
        end
    end

    def self.put_range(options, start_revision, end_revision)
        value = ""
        if end_revision
            value = "#{end_revision}..#{start_revision}"
        else
            value = "#{start_revision}"
        end

        if options.export
            puts value
        else
            puts "##teamcity[setParameter name='env.TEAMCITY_COMMIT_RANGE' value='#{value}']"
        end
    end

    def self.get_last_successful_rev(options, branch)
        STDERR.puts "Getting last successful revision from branch #{branch}"
        auth = "#{options.username}:#{options.password}"
        auth_token = Base64.strict_encode64(auth)

        headers = { "Authorization" => "Basic #{auth_token}",
                    "Accept" => "application/json" 
                  }
        
        end_revision = nil
        escaped_branch = URI.escape(branch)
        
        full_url = "#{options.url}/app/rest/buildTypes/id:#{options.build_type_id}/builds/?locator=status:SUCCESS,branch:name:#{escaped_branch}"
        response = RestClient.get(full_url, headers=headers)
        json_response = JSON.parse(response.body)

        if json_response["build"] && json_response["build"].count > 0
            last_build = json_response["build"].first
            last_build_id = last_build["id"]
            STDERR.puts "Got #{json_response["build"].count} successful builds, last successful was #{last_build_id}"
            

            build_full_url = "#{options.url}/app/rest/builds/id:#{last_build["id"]}"
            build_response = RestClient.get(build_full_url, headers=headers)
            build_json_response = JSON.parse(build_response.body) 

            end_revision = build_json_response["revisions"]["revision"][0]["version"]
        end
        
        return end_revision
    end

    def self.fetch_commit_range(options)
        branch = `git rev-parse --abbrev-ref HEAD`
        branch.strip!
        
        start_revision = ENV["BUILD_VCS_NUMBER"]
        STDERR.puts "Start revision is #{start_revision}"
        
        begin
            check_branches = [branch, "refs/heads/#{branch}", branch.split("/").last]
            end_revision = nil
            while end_revision == nil && check_branches.length > 0 do
                check_branch = check_branches.shift
                end_revision = get_last_successful_rev(options, check_branch)
            end
            
            put_range(options, start_revision, end_revision)
        rescue RestClient::Exception => e
            puts "ERROR getting last successful build from TeamCity:"
            pp e
            exit 1
        end 
    end

    def self.print_help()
        OptionsParser.parse(["--help"])
    end

    def self.commandline_fetch_commit_range(args = nil)
        options = OptionsParser.parse(args)

        fetch_commit_range(options)
    end
end

if __FILE__ == $0
    WTBuildHelpers::TeamCity.commandline_fetch_commit_range(ARGV)
end